from numbaml.dbml import RidgeCausalInference
from sklearn.linear_model import LinearRegression
from sklearn.datasets import make_regression
from numpy.typing import NDArray
import numpy as np
import unittest


class TestRegression(unittest.TestCase):

    n_tests = 5
    atol = 1e-06
    rtol = 1e-06
    min_samples = 15
    max_samples = 10_000

    def create_data(self, n_treatments: int = 1) -> tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]:
        n_samples = np.random.randint(self.min_samples, self.max_samples)
        max_features = 100 if n_samples // 2 > 100 else n_samples // 2
        n_features = np.random.randint(1, max_features)
        noise = np.random.randint(10, 100)

        x, y = make_regression(n_samples=n_samples, n_features=n_features, noise=noise)
        t = np.random.randint(0, 2, n_samples * n_treatments).reshape(n_samples, n_treatments)
        random_noise = np.random.normal(1, .1, n_samples * n_treatments).reshape(n_samples, n_treatments)
        y += (random_noise * t).sum(axis=1)

        return x, t, y

    def run_test(self, x, t, y):
        rci = RidgeCausalInference(x, t, y, alphas=[0], out_of_fold_residuals=False)
        rci.run_causal_pipeline()

        m = LinearRegression()
        m.fit(np.concatenate((x, t), axis=1), y)

        np.testing.assert_allclose(rci.treatment_coef_, m.coef_[-t.shape[1]:], atol=self.atol, rtol=self.rtol)

    def test_fwl_theorem(self):
        """alpha=0 and out-of-fold residuals oof=False should yield the same result according to FWL as OLS"""
        print('testing FWL')
        for i in range(self.n_tests):
            print(f'    test {i + 1}')
            x, t, y = self.create_data()
            self.run_test(x, t, y)

    def test_multiple_test_treatments(self):
        """alpha=0 and out-of-fold residuals oof=False should yield the same result according to FWL as OLS"""
        print('testing multiple test treatments')
        for i in range(self.n_tests):
            print(f'    test {i + 1}')

            x, t, y = self.create_data(n_treatments=np.random.randint(2, 10))
            self.run_test(x, t, y)



from ..metrics import r2_score
from .predict import predict
from .fit import fit

from numpy.typing import NDArray
from typing import Optional
import numpy as np



class _BaseModel:
    """https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LinearRegression.html"""
    def __init__(self, alpha: Optional[float] = None, fit_intercept: bool = True):
        """
        :param fit_intercept: Whether to calculate the intercept for this model.
            If set to false, no intercept will be used in calculations (i.e. data is expected to be centered).
        """
        self.fit_intercept = fit_intercept
        self.n_features_in_: Optional[int] = None
        self.feature_names_in_: Optional[list[str]] = None
        self.coef_: Optional[NDArray] = None
        self.params_: Optional[NDArray] = None
        self.intercept_: Optional[float] = None
        self.alpha_: Optional[float] = alpha
        self._X: Optional[NDArray] = None
        self._y: Optional[NDArray] = None

    def fit(self, x, y):
        self._assign_feature_names(x)
        self._X, self._y = self._to_numpy(x), self._to_numpy(y).reshape(x.shape[0], -1)
        self._X = self._add_intercept(self._X)
        self.params_ = fit(self._X, self._y, l2_penalty=self.alpha_, fit_intercept=self.fit_intercept)
        self._assign_params()

    def predict(self, x) -> NDArray[np.float64]:
        x = self._to_numpy(x)
        x = self._add_intercept(x)
        preds = predict(x, self.params_)
        return preds.flatten() if preds.shape[1] == 1 else preds

    def _add_intercept(self, x: NDArray) -> NDArray[np.float64]:
        if self.fit_intercept:
            n_samples = x.shape[0]
            intercept = np.ones((n_samples, 1), dtype=np.float64)
            x = np.concatenate([intercept, x], axis=1)
        return x

    def _assign_params(self):
        self.intercept_ = self.params_[0] if self.fit_intercept else np.float64(0)
        coef_ = self.params_[1:] if self.fit_intercept else self.params_
        self.coef_ = coef_.flatten() if coef_.shape[1] == 1 else coef_.T

    def _assign_feature_names(self, x):
        self.n_features_in_ = x.shape[1]
        self.feature_names_in_ = list(x.columns) if hasattr(x, 'columns') else list(map(str, range(self.n_features_in_)))

    @staticmethod
    def _to_numpy(a) -> NDArray[np.float64]:
        return np.asarray(a, dtype=np.float64, order='C')

    def score(self, x, y) -> float:
        """
        :param x: test sample
        :param y: true values for x
        :return: coefficient of determination
        """
        y_pred = self.predict(x)
        return np.float64(r2_score(y.flatten(), y_pred.flatten()))



from ..linear_model import approximate_leave_one_out_errors
from ..linear_model import RidgeCV, LinearRegression

from numpy.typing import NDArray
from typing import Optional
import numpy as np


class RidgeCausalInference:
    """
    Causal inference with de-biased / double machine using principles from Frisch–Waugh–Lovell theorem

    Causal inference is done by decomposing the regression problem into three steps:
        1. de-noise: regress y on X (confounding variables) and take residuals
        2. de-bias: regress treatment variable on X (confounding variables) and take residuals
        3. inference: find coefficient of treatment by regressing residuals from step 1 on the residuals from step 2

    RidgeCV is used in de-noising & de-biasing steps to reduce likelihood over-fitting,
    for handling large number of features relative to samples as well as correlated features.
    Out-of-fold residuals are also preferred to reduce likelihood over-fitting.
    Standard OLS regression is used to find the best estimate for the treatment effect.
    """

    def __init__(self, x, t, y,
                 alphas: list[float] = (0.1, 1.0, 10.0),
                 sig_level: float = 0.05,
                 out_of_fold_residuals: bool = True
                 ):
        """
        :param x: matrix of confounding features
        :param t: vector indicating treatment which we want to measure
        :param y: target variable
        :param sig_level: probability of a type-ii error for calculating confidence intervals
        :param out_of_fold_residuals: whether to use out-of-fold residuals (highly advised) or standard residuals
        """
        self.alphas = alphas
        self.sig_level = sig_level
        self.out_of_fold_residuals = out_of_fold_residuals

        self._X, self._t, self._y = self._to_numpy(x), self._to_numpy(t), self._to_numpy(y)

        self.t_residuals: Optional[NDArray[np.float64]] = None
        self.y_residuals: Optional[NDArray[np.float64]] = None
        self.treatment_coef_: Optional[float] = None
        self.confidence_interval_: Optional[NDArray[np.float64]] = None

    @staticmethod
    def _to_numpy(a) -> NDArray[np.float64]:
        a = np.asarray(a, dtype=np.float64, order='C')
        return a.reshape(a.shape[0], -1)

    def _residuals(self, target: NDArray[np.float64]) -> NDArray[np.float64]:
        m = RidgeCV(self.alphas)
        m.fit(self._X, target)
        if self.out_of_fold_residuals:
            res = approximate_leave_one_out_errors(m._X, target, m.alpha_, fit_intercept=True)
            return res.reshape(target.shape[0], -1)
        else:
            pred = m.predict(self._X).reshape(target.shape[0], -1)
            return target - pred

    def denoise(self):
        self.y_residuals = self._residuals(self._y)

    def debias(self):
        self.t_residuals = self._residuals(self._t)

    def inference(self):
        m = LinearRegression(fit_intercept=False)
        m.fit(self.t_residuals, self.y_residuals)
        self.treatment_coef_ = m.coef_
        self.confidence_interval_ = m.conf_int(sig=self.sig_level)

    def run_causal_pipeline(self):
        self.denoise()
        self.debias()
        self.inference()



from ..metrics import sum_square_residuals
from .predict import predict

from numba import njit, int64, float64
from scipy import stats
import numpy as np


@njit(int64(float64[:, ::1]), cache=True)
def residual_degrees_of_freedom(x):
    """
    The residual degree of freedom defined as the number of observations minus
    the rank of the regressor matrix (linearly independent features)
    """
    n_samples = x.shape[0]
    rank = np.linalg.matrix_rank(x)
    return np.max(np.array([n_samples - rank, 1]))


@njit(float64[::1](float64[:, ::1], float64), cache=True)
def parameter_standard_errors(x, scale):
    """
    The variance/covariance matrix can be of a linear contrast of the estimated parameters or all params multiplied by
    scale which will usually be an estimate of sigma².
    Scale will typically be the mean square error from the estimated model (sigma²).
    - off-diagonal elements, Cij, represent the covariance between the ith and jth estimated regression coefficients
    - symmetrical elements Cjj represent variance of the jth regression coefficient
    """
    c = np.linalg.inv(x.T @ x) * scale
    v = np.diag(c)
    return np.sqrt(v)


@njit(float64[:, ::1](float64[:, ::1], float64[:, ::1], float64[:, ::1], int64, float64), cache=True)
def parameter_confidence_intervals(x, y, weights, dof, t_value):
    """
    https://www.statsmodels.org/stable/generated/statsmodels.regression.linear_model.OLSResults.conf_int.html
    https://www.statsmodels.org/stable/_modules/statsmodels/regression/linear_model.html
    https://github.com/statsmodels/statsmodels/blob/main/statsmodels/base/model.py
    lower, upper = params - q * bse, params + q * bse
    """
    y_fitted = predict(x, weights)
    mse = sum_square_residuals(y.flatten(), y_fitted.flatten()) / dof
    se = parameter_standard_errors(x, mse)
    gap = t_value * se
    return np.stack((weights.flatten() - gap, weights.flatten() + gap))


def calculate_t_value(sig, dof):
    return stats.t.ppf(1 - sig / 2, dof)


def conf_int_parameter_method(x, y, params, sig):
    dof = residual_degrees_of_freedom(x)
    t_value = calculate_t_value(sig=sig, dof=dof)
    return parameter_confidence_intervals(x, y, params, dof, t_value).T



from sklearn.datasets import make_regression
import statsmodels.api as sm
import numpy.typing as npt
from typing import Tuple
import numpy as np
import unittest
import numbaml.linear_model
import sklearn.linear_model


class TestRegression(unittest.TestCase):

    n_tests = 5
    places = 6
    atol = 1e-06
    rtol = 1e-06
    min_samples = 15
    max_samples = 10_000

    def common_tests(self, model1, model2, x, y):
        print('        test .preds')
        np.testing.assert_allclose(model1.predict(x), model2.predict(x), atol=self.atol, rtol=self.rtol)
        print('        test .coef_')
        np.testing.assert_allclose(model1.coef_, model2.coef_, atol=self.atol, rtol=self.rtol)
        print('        test .intercept_')
        np.testing.assert_allclose(model1.intercept_, model2.intercept_, atol=self.atol, rtol=self.rtol)
        print('        test .n_features_in_')
        np.testing.assert_allclose(model1.n_features_in_, model2.n_features_in_, atol=self.atol, rtol=self.rtol)

    def ci_test(self, model, x, y, fit_intercept):
        print('        test .conf_int')
        x = sm.add_constant(x) if fit_intercept else x
        m3 = sm.OLS(y, x)
        r3 = m3.fit()
        sig = np.random.choice([0.01, 0.025, 0.05, 0.1], size=None)
        ci1 = model.conf_int(sig=sig)
        ci3 = r3.conf_int(alpha=sig)
        np.testing.assert_allclose(ci1, ci3, atol=self.atol, rtol=self.rtol)

    def alpha_test(self, model1, model2):
        print('        test .alpha_')
        self.assertEqual(model1.alpha_, model2.alpha_)

    def score_test(self, model1, model2, x, y):
        print('        test .score')
        self.assertAlmostEqual(model1.score(x, y), model2.score(x, y), places=self.places)

    def best_score_test(self, model1, model2):
        print('        test .best_score_')
        self.assertAlmostEqual(model1.best_score_, model2.best_score_, places=self.places)

    @staticmethod
    def select_scoring(cv: int) -> str:
        scoring = None if cv is None else np.random.choice(['r2', 'neg_mean_squared_error'])
        return scoring

    def select_cv(self) -> int:
        leave_one_out = np.random.choice([True, False])
        cv = np.random.randint(2, self.min_samples) if leave_one_out else None
        return cv

    @staticmethod
    def select_fit_intercept() -> bool:
        return np.random.choice([True, False])

    def create_data(self, n_targets: int = 1) -> Tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
        n_samples = np.random.randint(self.min_samples, self.max_samples)
        max_features = 100 if n_samples // 2 > 100 else n_samples // 2
        n_features = np.random.randint(1, max_features)
        noise = np.random.randint(10, 100)
        return make_regression(n_samples=n_samples, n_features=n_features, noise=noise, n_targets=n_targets)

    @staticmethod
    def select_alphas(n_alphas: int) -> npt.NDArray[np.float64]:
        alphas = np.array([.001, .01, .1, .5, .9, .99, 1., 1.1, 2.5, 5., 10.])
        alphas = np.random.choice(alphas, n_alphas, replace=False)
        return alphas

    @staticmethod
    def fit_models(x, y, model_type1, model_type2, kwargs: dict) -> Tuple:
        m1 = model_type1(**kwargs)
        m2 = model_type2(**kwargs)
        m1.fit(x, y)
        m2.fit(x, y)
        return m1, m2

    def test_linear_regression(self):
        print('testing LinearRegression')
        for i in range(self.n_tests):
            print(f'    test {i + 1}')
            x, y = self.create_data()
            kwargs = {'fit_intercept': self.select_fit_intercept()}
            m1, m2 = self.fit_models(x, y,
                                     numbaml.linear_model.LinearRegression,
                                     sklearn.linear_model.LinearRegression,
                                     kwargs=kwargs
                                     )
            self.common_tests(m1, m2, x, y)
            self.score_test(m1, m2, x, y)
            self.ci_test(m1, x, y, kwargs['fit_intercept'])

    def test_ridge(self):
        print('testing Ridge')
        for i in range(self.n_tests):
            print(f'    test {i + 1}')
            x, y = self.create_data()
            kwargs = {'alpha': self.select_alphas(1), 'fit_intercept': self.select_fit_intercept()}
            m1, m2 = self.fit_models(x, y,
                                     numbaml.linear_model.Ridge,
                                     sklearn.linear_model.Ridge,
                                     kwargs=kwargs
                                     )
            self.common_tests(m1, m2, x, y)
            self.score_test(m1, m2, x, y)

    def test_ridgecv(self):
        print('testing RidgeCV')
        for i in range(self.n_tests):
            print(f'    test {i + 1}')
            x, y = self.create_data()
            cv = self.select_cv()
            kwargs = {'alphas': self.select_alphas(np.random.randint(2, 10)),
                      'scoring': self.select_scoring(cv),
                      'cv': cv,
                      'fit_intercept': self.select_fit_intercept()
                      }
            m1, m2 = self.fit_models(x, y,
                                     numbaml.linear_model.RidgeCV,
                                     sklearn.linear_model.RidgeCV,
                                     kwargs=kwargs
                                     )
            self.common_tests(m1, m2, x, y)
            self.score_test(m1, m2, x, y)
            self.alpha_test(m1, m2)
            self.best_score_test(m1, m2)

    def test_multivariate(self):
        print('testing multivariate')
        for i in range(self.n_tests):
            print(f'    test {i + 1}')
            x, y = self.create_data(n_targets=np.random.randint(2, 10))
            kwargs = {'fit_intercept': self.select_fit_intercept()}
            m1, m2 = self.fit_models(x, y,
                                     numbaml.linear_model.LinearRegression,
                                     sklearn.linear_model.LinearRegression,
                                     kwargs=kwargs
                                     )
            self.common_tests(m1, m2, x, y)



from .confidence_intervals import conf_int_parameter_method
from .base_model import _BaseModel

from numpy.typing import NDArray
import numpy as np


class LinearRegression(_BaseModel):
    """https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LinearRegression.html"""
    def __init__(self, fit_intercept: bool = True):
        """
        :param fit_intercept: Whether to calculate the intercept for this model.
            If set to false, no intercept will be used in calculations (i.e. data is expected to be centered).
        """
        super().__init__(alpha=0, fit_intercept=fit_intercept)

    def conf_int(self, sig: float = .05) -> NDArray[np.float64]:
        if self._y.shape[1] > 1:
            raise NotImplementedError('Confidence intervals are only implemented for single targets')
        return conf_int_parameter_method(self._X, self._y, self.params_, sig)

    def conf_int_dict(self, sig: float = .05) -> dict:
        conf_int = self.conf_int(sig).T
        names = (['intercept'] + self.feature_names_in_) if self.fit_intercept else self.feature_names_in_
        return {
            'feature_name': names,
            'lower_bound': conf_int[0],
            'coef': self.params_,
            'upper_bound': conf_int[1]
        }



from numba import njit, float64, boolean
import numpy as np


@njit(float64(float64[::1], float64[::1]), cache=True)
def sum_square_residuals(y_true, y_pred):
    return np.sum((y_true - y_pred) ** 2)


@njit(float64(float64[::1], float64[::1], boolean), cache=True)
def mean_squared_error(y_true, y_pred, squared: bool = True):
    """https://scikit-learn.org/stable/modules/generated/sklearn.metrics.mean_squared_error.html"""
    mse = sum_square_residuals(y_true, y_pred) / y_true.size
    return mse if squared else mse ** .5


@njit(float64(float64[::1], float64[::1]), cache=True)
def neg_mean_squared_error(y_true, y_pred):
    return - mean_squared_error(y_true, y_pred, squared=True)


@njit(float64(float64[::1], float64[::1]), cache=True)
def r2_score(y_true, y_pred):
    """https://scikit-learn.org/stable/modules/generated/sklearn.metrics.r2_score.html#sklearn.metrics.r2_score"""
    y_mean = np.mean(y_true)
    total_sum_squares = np.sum((y_true - y_mean) ** 2)
    residual_sum_squares = sum_square_residuals(y_true, y_pred)
    return 1 - (residual_sum_squares / total_sum_squares)



from .linear_model import approximate_leave_one_out_errors
from .linear_model import LinearRegression
from .linear_model import Ridge, RidgeCV

from .dbml import RidgeCausalInference

from .metrics import (
mean_squared_error,
neg_mean_squared_error,
sum_square_residuals,
r2_score
)


from ..metrics import r2_score, neg_mean_squared_error
from .fit import fit, create_penalty_matrix
from .predict import predict

from numba import njit, float64, int64, types, prange, boolean
import numpy as np


@njit(boolean[:, ::1](int64, int64), cache=True)
def kfold(n_samples, n_splits):
    """scikit-learn.org/stable/modules/generated/sklearn.model_selection.KFold.html#sklearn.model_selection.KFold"""
    first_splits = n_samples % n_splits
    test_flags = np.zeros(shape=(n_splits, n_samples), dtype=boolean)
    start = 0
    for i in range(n_splits):
        size = n_samples // n_splits + 1 if i < first_splits else n_samples // n_splits
        end = start + size
        test_flags[i][start: end] = np.ones(size, dtype=boolean)
        start += size
    return test_flags


@njit(float64(float64[::1], float64[::1], boolean), cache=True)
def calculate_score(y_true, y_pred, r2):
    return r2_score(y_true, y_pred) if r2 else neg_mean_squared_error(y_true, y_pred)


@njit(types.UniTuple(float64, 2)(float64[:, ::1], float64[:, ::1], float64[::1], boolean, int64, boolean),
      parallel=True, cache=True)
def find_alpha_kfolds(x, y, alphas, fit_intercept, cv, r2):
    n_samples, n_features = x.shape

    best_score = -np.inf
    best_alpha = np.float64(0)

    all_test_flags = kfold(n_samples, n_splits=cv)

    for i in range(alphas.size):
        a = alphas[i]
        scores = np.zeros(cv, dtype=np.float64)
        for j in prange(cv):

            is_test = all_test_flags[j]
            test_x, test_y = x[is_test], y[is_test]
            train_x, train_y = x[~is_test], y[~is_test]

            weights = fit(train_x, train_y, a, fit_intercept)
            preds = predict(test_x, weights)

            scores[j] = calculate_score(test_y.flatten(), preds.flatten(), r2)
        avg_score = np.mean(scores)
        if avg_score > best_score:
            best_score = avg_score
            best_alpha = a

    return best_alpha, best_score


@njit(float64[::1](float64[:, ::1], float64[:, ::1], float64, boolean), cache=True)
def approximate_leave_one_out_errors(x, y, l2_penalty, fit_intercept):
    """https://medium.com/@jcatankard_76170/efficient-leave-one-out-cross-validation-f1dee3b68dfe"""

    weights = fit(x, y, l2_penalty, fit_intercept)
    preds = predict(x, weights)
    residuals = y - preds

    penalty = create_penalty_matrix(l2_penalty, x.shape[1], fit_intercept)

    h = np.diag(x @ np.linalg.inv(x.T @ x + penalty) @ x.T)
    return (residuals / (1 - h).reshape(-1, 1)).flatten()


@njit(types.UniTuple(float64, 2)(float64[:, ::1], float64[:, ::1], float64[::1], boolean), parallel=True, cache=True)
def find_alpha_loo(x, y, alphas, fit_intercept):

    all_scores = np.zeros(alphas.size, dtype=np.float64)
    for i in prange(alphas.size):
        errors = approximate_leave_one_out_errors(x, y, alphas[i], fit_intercept)
        # calculate neg mean squared error
        all_scores[i] = - np.mean(errors ** 2)

    best_alpha = alphas[np.argmax(all_scores)]
    best_score = np.max(all_scores)
    return best_alpha, best_score



from .model_selection import find_alpha_kfolds, find_alpha_loo
from .base_model import _BaseModel
from .fit import fit

from typing import Optional
import numpy as np


class Ridge(_BaseModel):
    """https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.Ridge.html#sklearn.linear_model.Ridge"""

    def __init__(self, alpha: float = 1., fit_intercept: bool = True):
        """
        :param alpha: l2 penalization.
        :param fit_intercept: Whether to calculate the intercept for this model.
            If set to false, no intercept will be used in calculations (i.e. data is expected to be centered).
        """
        super().__init__(alpha=float(alpha), fit_intercept=fit_intercept)


class RidgeCV(_BaseModel):
    """scikit-learn.org/stable/modules/generated/sklearn.numbaml.RidgeCV.html#sklearn.numbaml.RidgeCV"""

    def __init__(self, alphas: list[float] = (0.1, 1.0, 10.0), cv: Optional[int] = None, scoring: Optional[str] = None,
                 fit_intercept: bool = True):
        """
        :param alphas: array-like of shape (n_alphas,), default=(0.1, 1.0, 10.0)
            Regularization strength; must be a positive float.
            Regularization improves the conditioning of the problem and reduces the variance of the estimates.
            Larger values specify stronger regularization
        :param cv: Determines the cross-validation splitting strategy
            None, to use the efficient Leave-One-Out cross-validation
            integer, to specify the number of folds.
        :param scoring: If None, the negative mean squared error if cv is None
            (i.e. when using leave-one-out cross-validation) and r2 score otherwise.
            'r2' for (coefficient of determination) regression score
            'neg_mean_squared_error' for the negative mean squared error
        :param fit_intercept: Whether to calculate the intercept for this model.
            If set to false, no intercept will be used in calculations (i.e. data is expected to be centered).
        """
        super().__init__(fit_intercept=fit_intercept)
        self.alphas = np.array(alphas, dtype=np.float64)
        self.cv = cv
        self.scoring = scoring
        self.r2: Optional[bool] = None
        self.gcv: Optional[str] = None
        self.best_score_: Optional[float] = None

    def fit(self, x, y):
        self._assign_feature_names(x)
        self._assign_cv(x.shape[0])
        self._assign_scoring(x.shape[0])

        self._X, self._y = self._to_numpy(x), self._to_numpy(y).reshape(x.shape[0], -1)
        self._X = self._add_intercept(self._X)

        if self.gcv == 'svd':
            self.alpha_, self.best_score_ = find_alpha_loo(self._X, self._y, self.alphas, self.fit_intercept)
        else:
            self.alpha_, self.best_score_ = find_alpha_kfolds(self._X, self._y, self.alphas, self.fit_intercept,
                                                              self.cv, self.r2)

        self.params_ = fit(self._X, self._y, self.alpha_, self.fit_intercept)
        self._assign_params()

    def _assign_cv(self, n_samples: int):
        self.cv = n_samples if self.cv is None else self.cv
        if (self.cv <= 1) | (self.cv > n_samples):
            raise ValueError('CV must be >= 2 and <= number of samples')
        self.gcv = 'svd' if (self.cv == n_samples) & (n_samples > self.n_features_in_) else 'eigen'

    def _assign_scoring(self, n_samples: int):
        if self.scoring is None:
            self.scoring = 'r2' if self.cv < n_samples else 'neg_mean_squared_error'
        elif (self.scoring == 'r2') & (self.cv == n_samples):
            raise ValueError('R^2 score is not well-defined with leave one out cv.')
        self.r2 = self.scoring == 'r2'



from numba import njit, int64, float64, boolean
import numpy as np


@njit(float64[:, ::1](float64, int64, boolean), cache=True)
def create_penalty_matrix(l2_penalty, n_features, fit_intercept):
    """n_features includes intercept"""
    identity = np.identity(n_features)
    # set top corner to zero as we don't penalize intercept
    identity[0][0] = 0 if fit_intercept else 1
    return l2_penalty * identity


@njit(float64[:, ::1](float64[:, ::1], float64[:, ::1], float64, boolean), cache=True)
def fit(x, y, l2_penalty, fit_intercept):
    """
    Solution 1 (solve regression formula for B)
    solve for B: y = X.B -> Xt.y = (Xt.X).B -> Xt.(Xt.X)^-1.y = (Xt.X).(Xt.X)^-1.B = I.B = B
    add penalty: Xt.(Xt.X + A)^-1.y (we don't penalize intercept, see code)

    Solution 2 (minimise SSE) web.stanford.edu/~mrosenfe/soc_meth_proj3/matrix_OLS_NYU_notes.pdf
    error = ypi - yi = b0 + bi.xi + e - yi
    SSS = E(b0 + bi.xi - yi)^2
    SSE = (matrix: intercept->B) (B.X - y)^2 = (B.X - y).(B.X - y)t = B.X.Bt.Xt - y.Bt.Xt - yt.B.X + y.yt
    = yt.y + B.X.Bt.Xt - 2.y.Bt.Xt (y.Bt.Xt == yt.B.X)
    dSSE/db = 2.B.X.Xt - 2.y.Xt = -2.Xt.(y - B.X)
    minimise SSE: 0 = -2.Xt.(y - B.X) = -Xt.y + B.X.Xt -> Xt.y = B.X.Xt -> (X.Xt)^-1.Xt.y = B.(X.Xt).(X.Xt)^-1 = B

    :param x: independent variables
    :param y: dependent target
    :param l2_penalty: regularization penalty
    :param fit_intercept: if l2 penalty is 0, this has no impact
    :return: coefficients, intercept
    """
    penalty = create_penalty_matrix(l2_penalty, x.shape[1], fit_intercept)
    weights = np.linalg.inv(x.T @ x + penalty) @ x.T @ y
    return weights.reshape(x.shape[1], y.shape[1])
